{**************************************************************************************}
{                                                                                      }
{ CCR Exif - Delphi class library for reading and writing Exif metadata in JPEG files  }
{ Version 1.1.2 (2011-01-23)                                                           }
{                                                                                      }
{ The contents of this file are subject to the Mozilla Public License Version 1.1      }
{ (the "License"); you may not use this file except in compliance with the License.    }
{ You may obtain a copy of the License at http://www.mozilla.org/MPL/                  }
{                                                                                      }
{ Software distributed under the License is distributed on an "AS IS" basis, WITHOUT   }
{ WARRANTY OF ANY KIND, either express or implied. See the License for the specific    }
{ language governing rights and limitations under the License.                         }
{                                                                                      }
{ The Original Code is ccr_exif_streamhelper.pas.                                      }
{                                                                                      }
{ The Initial Developer of the Original Code is Chris Rolliston. Portions created by   }
{ Chris Rolliston are Copyright (C) 2009-2011 Chris Rolliston. All Rights Reserved.    }
{                                                                                      }
{**************************************************************************************}

unit ccr_exif_streamhelper;

{$mode delphi}{$H+}
{$ASMMODE intel}

interface

uses
  SysUtils, Classes;

type

  TSuperFindStringList = Class(TStringList)
  public
    function Find(const S: string; Out Index: Integer): Boolean; override;
    function IndexOf(const S: string): Integer; override;
  end;

  { TUnicodeStringList }

  TUnicodeStringList = TSuperFindStringList;
//  TUnicodeStringList = TStringList;

  TEndianness = (SmallEndian, BigEndian);

  TStreamHelper = class helper for TStream
    function ReadByte: Byte; overload;
    function ReadSmallInt(Endianness: TEndianness): SmallInt; overload;
    function ReadWord(Endianness: TEndianness): Word; overload;
    function ReadLongInt(Endianness: TEndianness): LongInt; overload;
    function ReadLongWord(Endianness: TEndianness): LongWord; overload;
    function ReadDouble(Endianness: TEndianness): Double;
    function ReadShortString(var S: ShortString): Boolean; overload;
    function ReadShortString: ShortString; overload;
    function TryReadBuffer(var Buffer; Count: Integer): Boolean;
    function TryReadHeader(const Header; HeaderSize: Byte): Boolean;
    procedure WriteByte(Value: Byte); overload;
    procedure WriteByte(Value: AnsiChar); overload;
    procedure WriteWord(Value: Word; Endianness: TEndianness);
    procedure WriteSmallInt(Value: SmallInt; Endianness: TEndianness);
    procedure WriteLongInt(Value: LongInt; Endianness: TEndianness);
    procedure WriteLongWord(Value: LongWord; Endianness: TEndianness);
    procedure WriteDouble(Value: Double; Endianness: TEndianness);
    procedure WriteUTF8Chars(const S: UTF8String); overload;
    procedure WriteUTF8Chars(const S: UnicodeString); overload; 
    procedure WriteUTF8Chars(const S: UnicodeString; const Args: array of const); overload;
    procedure WriteWideChars(const S: UnicodeString; Endianness: TEndianness);
  end;

  TUserMemoryStream = class(TCustomMemoryStream) //read-only stream access on an existing buffer;
  public
    constructor Create(Memory: Pointer; Size: Integer);
    procedure ChangeMemory(NewMemory: Pointer; NewSize: Integer);
    function Write(const Buffer; Count: Integer): Integer; override;
  end;

function SwapLongInt(const Value: LongInt): LongInt;
function SwapLongWord(const Value: LongWord): LongWord;
function SwapSingle(const Value: Single): Single;
function SwapDouble(const Value: Double): Double;

function UnicodeFormat(const S: UnicodeString; const Args: array of const): UnicodeString;
function UnicodeSameText(const S1, S2: String): Boolean; inline;

implementation

uses RTLConsts, SysConst;

function SwapLongInt(const Value: LongInt): LongInt;
asm
  BSWAP EAX
end;

function SwapLongWord(const Value: LongWord): LongWord;
asm
  BSWAP EAX
end;

function SwapSingle(const Value: Single): Single;
asm
  BSWAP EAX
end;

function SwapDouble(const Value: Double): Double;
var
  I: Integer;
begin
  for I := SizeOf(Double) - 1 downto 0 do
    PByteArray(@Result)[I] := PByteArray(@Value)[SizeOf(Double) - 1 - I];
end;

function UnicodeFormat(const S: UnicodeString; const Args: array of const): UnicodeString;
begin
  Result := Format(S, Args);
end;

function UnicodeSameText(const S1, S2: String): Boolean;
begin
  Result := SameText(S1, S2);
end;

{ TSuperFindStringList }

function TSuperFindStringList.Find(const S: string; out Index: Integer): Boolean;
var
  new_s: String;
begin
  Result := inherited Find(S, Index);
  if not Result then
    if (Copy(S, Length(S), Length('/'))<>'/') and
       (Copy(S, Length(S), Length('#'))<>'#') then
    begin
      new_s := S+'/';
      Result := inherited Find(new_s, Index);
    end;
  if not Result then
    if (Copy(S, Length(S), Length('/'))='/') then
    begin
      new_s := Copy(S, 1, Length(S)-Length('/'));
      Result := inherited Find(new_s, Index);
    end;
end;

function TSuperFindStringList.IndexOf(const S: string): Integer;
begin
  If Not Find (S,Result) then
    Result:=-1;
end;

{ TStreamHelper }

function TStreamHelper.ReadByte: Byte;
begin
  ReadBuffer(Result, 1);
end;

function TStreamHelper.ReadWord(Endianness: TEndianness): Word;
begin
  if Endianness = BigEndian then
    Result := BEtoN(readword)
  else
    Result := LEtoN(readword);
end;

function TStreamHelper.ReadDouble(Endianness: TEndianness): Double;
begin
  ReadBuffer(Result, 8);
  if Endianness = BigEndian then
    Result := SwapDouble(Result);
end;

function TStreamHelper.ReadLongInt(Endianness: TEndianness): LongInt;
begin
  if Endianness = BigEndian then
    Result := BEtoN(readdword)
  else
    Result := LEtoN(readdword);
end;

function TStreamHelper.ReadLongWord(Endianness: TEndianness): LongWord;
begin
  if Endianness = BigEndian then
    Result := BEtoN(readdword)
  else
    Result := LEtoN(readdword);
end;

function TStreamHelper.ReadShortString(var S: ShortString): Boolean;
begin
  Result := (Read(S[0], 1) = 1) and (Read(S[1], Byte(S[0])) = Byte(S[0]));
end;

function TStreamHelper.ReadShortString: ShortString;
begin
  if not ReadShortString(Result) then
    raise EReadError.CreateRes(@SReadError);
end;

function TStreamHelper.ReadSmallInt(Endianness: TEndianness): SmallInt;
begin
  if Endianness = BigEndian then
    Result := BEtoN(readword)
  else
    Result := LEtoN(readword);
end;

function TStreamHelper.TryReadBuffer(var Buffer; Count: Integer): Boolean;
begin
  Result := (Read(Buffer, Count) = Count)
end;

function TStreamHelper.TryReadHeader(const Header; HeaderSize: Byte): Boolean;
var
  Buffer: array[Byte] of Byte;
  BytesRead: Int64;
begin
  BytesRead := Read(Buffer, HeaderSize);
  Result := (BytesRead = HeaderSize) and CompareMem(@Buffer, @Header, HeaderSize);
  if not Result then Seek(-BytesRead, soCurrent);
end;

procedure TStreamHelper.WriteByte(Value: Byte);
begin
  WriteBuffer(Value, 1);
end;

procedure TStreamHelper.WriteByte(Value: AnsiChar);
begin
  WriteBuffer(Value, 1);
end;

procedure TStreamHelper.WriteDouble(Value: Double; Endianness: TEndianness);
begin
  if Endianness = BigEndian then
    Value := SwapDouble(Value);
  WriteBuffer(Value, 8);
end;

procedure TStreamHelper.WriteWideChars(const S: UnicodeString; Endianness: TEndianness);
var
  Ch: WideChar;
begin
  if Endianness = SmallEndian then
    WriteBuffer(Pointer(S)^, Length(S) * 2)
  else
    for Ch in S do
    begin
      WriteBuffer(WordRec(Ch).Hi, 1);
      WriteBuffer(WordRec(Ch).Lo, 1);
    end;
end;

procedure TStreamHelper.WriteWord(Value: Word; Endianness: TEndianness);
begin
  if Endianness = BigEndian then
    inherited WriteWord(NtoBE(Value))
  else
    inherited WriteWord(NtoLE(Value));
end;

procedure TStreamHelper.WriteLongWord(Value: LongWord; Endianness: TEndianness);
begin
  if Endianness = BigEndian then
    WriteDWord(NtoBE(Value))
  else
    WriteDWord(NtoLE(Value));
end;

procedure TStreamHelper.WriteSmallInt(Value: SmallInt; Endianness: TEndianness);
begin
  if Endianness = BigEndian then
    inherited WriteWord(Word(NtoBE(Value)))
  else
    inherited WriteWord(Word(NtoLE(Value)));
end;

procedure TStreamHelper.WriteLongInt(Value: LongInt; Endianness: TEndianness);
begin
  if Endianness = BigEndian then
    WriteDWord(Cardinal(NtoBE(Value)))
  else
    WriteDWord(Cardinal(NtoLE(Value)));
end;

procedure TStreamHelper.WriteUTF8Chars(const S: UTF8String);
begin
  WriteBuffer(Pointer(S)^, Length(S));
end;

procedure TStreamHelper.WriteUTF8Chars(const S: UnicodeString);
begin
  WriteUTF8Chars(UTF8Encode(S));
end;

procedure TStreamHelper.WriteUTF8Chars(const S: UnicodeString;
  const Args: array of const);
begin
  WriteUTF8Chars(UTF8Encode(UnicodeFormat(S, Args)));
end;

{ TUserMemoryStream }

constructor TUserMemoryStream.Create(Memory: Pointer; Size: Integer);
begin
  inherited Create;
  SetPointer(Memory, Size);
end;

procedure TUserMemoryStream.ChangeMemory(NewMemory: Pointer; NewSize: Integer);
begin
  SetPointer(NewMemory, NewSize);
  Position := 0;
end;

function TUserMemoryStream.Write(const Buffer; Count: Integer): Integer;
begin
  Result := 0;
end;

end.
